# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.Queue import Queue, QueueItem
from nyawc.http.Handler import Handler
from nyawc.http.Request import Request
from nyawc.helpers.HTTPRequestHelper import HTTPRequestHelper

import threading
import time

class Crawler:
    """The main Crawler class which handles the crawling recursion, queue and processes.

    Attributes:
        __options (obj): The options to use for the current crawling runtime.
        __queue (obj): The request/response pair queue containing everything to crawl.
        __stopping (bool): If the crawler is topping the crawling process.
        __stopped (bool): If the crawler finished stopping the crawler process.
        __lock (obj): The callback lock to prevent race conditions.

    """

    __options = None

    __queue = Queue()

    __stopping = False

    __stopped = False

    __lock = threading.Lock()

    def __init__(self, options):
        """Constructs a Crawler class.

        Args:
            options (obj): The options to use for the current crawling runtime.

        """

        self.__options = options

    def start_with(self, request):
        """Start the crawler using the given request.

        Args:
            request (obj): The startpoint for the crawler.

        """

        HTTPRequestHelper.patch_with_options(request, self.__options)
        self.__queue.add_request(request)

        self.__crawler_start()

    def __spawn_new_requests(self):
        """Spawn new requests until the max processes option value is reached."""

        concurrent_requests_count = self.__queue.get_count_including([QueueItem.STATUS_IN_PROGRESS])
        new_requests_spawned = False

        while concurrent_requests_count < self.__options.performance.max_threads:
            if self.__spawn_new_request():
                new_requests_spawned = True
                concurrent_requests_count += 1
            else:
                break

        if concurrent_requests_count == 0 and not new_requests_spawned and not self.__stopping:
            self.__crawler_stop()

    def __spawn_new_request(self):
        """Spawn the first queued request if available.

        Returns:
            bool: If a new request was spawned.

        """

        first_in_line = self.__queue.get_first(QueueItem.STATUS_QUEUED)
        if first_in_line is None:
            return False

        self.__request_start(first_in_line)
        return True

    def __crawler_start(self):
        """Spawn the first X queued request, where X is the max processes option."""

        self.__options.callbacks.crawler_before_start()

        try:
            self.__spawn_new_requests()

            while not self.__stopped: 
                time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            pass

    def __crawler_stop(self, force_quit=False):
        """Mark the crawler as stopped.

        Note:
            If self.__stopped is True, the main thread will be stopped. Every piece of code that gets
            executed after self.__stopped is True could cause Thread exceptions and or race conditions.

        Args:
            force_quit (bool): Also cancel any ongoing requests.

        """

        self.__stopping = True

        queued_items = self.__queue.get_all_including([
            QueueItem.STATUS_QUEUED, 
            QueueItem.STATUS_IN_PROGRESS
        ])

        for queue_item in queued_items:
            queue_item.status = QueueItem.STATUS_CANCELLED

        self.__crawler_finish()

        self.__stopped = True

    def __crawler_finish(self):
        """Called when the crawler is finished because there are no queued requests left or it was stopped."""

        self.__options.callbacks.crawler_after_finish(self.__queue)

    def __request_start(self, queue_item):
        """Execute the request in given queue item.

        Args:
            queue_item (obj): The request/response pair to scrape.

        """

        action = self.__options.callbacks.request_before_start(self.__queue, queue_item)

        if action == CrawlerActions.DO_STOP_CRAWLING:
            self.__crawler_stop(True)
            return

        if action == CrawlerActions.DO_SKIP_TO_NEXT:
            queue_item.status = QueueItem.STATUS_FINISHED
            return

        if action == CrawlerActions.DO_CONTINUE_CRAWLING or action is None:
            queue_item.status = QueueItem.STATUS_IN_PROGRESS

            thread = CrawlerThread(self.__request_finish, self.__lock, queue_item)
            thread.daemon = True
            thread.start()

    def __request_finish(self, queue_item, new_requests):
        """Called when the crawler finished the given queued item.

        Args:
            queue_item (obj): The request/response pair that finished.
            new_requests list(obj): All the requests that were found during this request.

        """

        new_queue_items = []
        action = None

        if queue_item.status not in [QueueItem.STATUS_ERRORED, QueueItem.STATUS_CANCELLED]:
            for new_request in new_requests:
                HTTPRequestHelper.patch_with_options(new_request, self.__options, queue_item.response)
                
                if HTTPRequestHelper.is_already_in_queue(new_request, self.__queue):
                    continue

                if not HTTPRequestHelper.complies_with_scope(self.__queue, queue_item, new_request, self.__options.scope):
                    continue

                new_request.depth = queue_item.request.depth + 1
                if self.__options.scope.max_depth is not None:
                    if new_request.depth > self.__options.scope.max_depth:
                        continue

                new_queue_item = self.__queue.add_request(new_request)
                new_queue_items.append(new_queue_item)

            queue_item.status = QueueItem.STATUS_FINISHED
            action = self.__options.callbacks.request_after_finish(self.__queue, queue_item, new_queue_items)

        if self.__stopping:
            return

        if action == CrawlerActions.DO_STOP_CRAWLING:
            self.__crawler_stop()
            return

        if action == CrawlerActions.DO_CONTINUE_CRAWLING or action is None:
            self.__spawn_new_requests()
            return

class CrawlerThread(threading.Thread):
    """The crawler thread executes the HTTP request using the HTTP handler.

    Attributes:
        __callback (obj): The method to call when finished
        __callback_lock (bool): The callback lock that prevents race conditions.
        __queue_item (obj): The queue item containing a request to execute.

    """

    __callback = None

    __callback_lock = None

    __queue_item = None

    def __init__(self, callback, callback_lock, queue_item):
        """Constructs a crawler thread class

        Args:
            callback (obj): The method to call when finished
            callback_lock (bool): The callback lock that prevents race conditions.
            queue_item (obj): The queue item containing a request to execute.

        """

        threading.Thread.__init__(self)
        self.__callback = callback
        self.__queue_item = queue_item
        self.__callback_lock = callback_lock

    def run(self):
        """Executes the HTTP call.

        Note:
            If the this and the parent handler raised an error, the queue item status will be set to errored
            instead of finished.

        """

        new_requests = []
            
        try:
            handler = Handler(self.__queue_item)
            new_requests = handler.get_new_requests()

            try:
                self.__queue_item.response.raise_for_status()
            except Exception:
                if self.__queue_item.request.parent_raised_error:
                    self.__queue_item.status = QueueItem.STATUS_ERRORED
                else:
                    for new_request in new_requests:
                        new_request.parent_raised_error = True

        except Exception as e:
            self.__queue_item.status = QueueItem.STATUS_ERRORED


        for new_request in new_requests:
            new_request.parent_url = self.__queue_item.request.url

        with self.__callback_lock:
            self.__callback(self.__queue_item, new_requests)

class CrawlerActions:
    """The actions that crawler callbacks can return.

    Attributes:
        DO_CONTINUE_CRAWLING (int): Continue by crawling the request.
        DO_SKIP_TO_NEXT (int): Skip the current request and continue with the next one in line.
        DO_STOP_CRAWLING (int): Stop crawling and quit ongoing requests.
        
    """

    DO_CONTINUE_CRAWLING = 1

    DO_SKIP_TO_NEXT = 2

    DO_STOP_CRAWLING = 3


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import requests
import importlib
import os

class Handler:
    """The Handler class executes HTTP requests.

    Attributes:
        __queue_item (obj): The queue item containing a request to execute.

    """

    __queue_item = None

    def __init__(self, queue_item):
        """Construct the HTTP handler.

        Args:
            queue_item (obj): The queue item containing the request.

        """

        self.__queue_item = queue_item

        self.__queue_item.response = self.__make_request(
            self.__queue_item.request.url,
            self.__queue_item.request.method,
            self.__queue_item.request.data,
            self.__queue_item.request.cookies,
            {
                'User-Agent': self.__queue_item.request.user_agent
            }
        )

    def get_new_requests(self):
        """Retrieve all the new request that were found in this request.

        Returns:
            list(obj): A list of request objects.

        """

        content_type = self.__queue_item.response.headers.get('content-type')
        scrapers = self.__get_all_scrapers()
        requests = []

        for scraper in scrapers:
            instance = scraper(self.__queue_item)
            if self.__content_type_matches(content_type, instance.content_types):
                requests.extend(instance.get_requests())

        return requests

    def __make_request(self, url, method, data, cookies, headers):
        """Execute a request with the given data.

        Args:
            url (str): The URL to call.
            method (str): The method (e.g. `get` or `post`).
            data (str): The data to call the URL with.
            cookies (obj): The cookie dict.
            headers (obj): The header dict.

        Returns:
            obj: The response object.

        """

        request_by_method = getattr(requests, method)
        return request_by_method(
            url=url, 
            data=data, 
            cookies=cookies, 
            headers=headers,
            allow_redirects=True,
            stream=True
        )

    def __get_all_scrapers(self):
        """Find all available scraper references.

        Returns:
            list(obj): The scraper references.

        """

        modules_strings = self.__get_all_scrapers_modules()
        modules = []

        for module_string in modules_strings:
            module = importlib.import_module("nyawc.scrapers." + module_string)
            modules.append(getattr(module, module_string))

        return modules

    def __get_all_scrapers_modules(self):
        """Find all available scraper modules.

        Returns:
            list(obj): The scraper modules.

        """

        modules = []
        
        file = os.path.realpath(__file__)
        folder = os.path.dirname(file)

        for filename in os.listdir(folder + "/../scrapers"):
            if filename.endswith("Scraper.py"):
                modules.append(filename[:-3])

        return modules

    def __content_type_matches(self, content_type, available_content_types):
        """Check if the given content type matches one of the available content types.

        Args:
            content_type (str): The given content type.
            available_content_types list(str): All the available content types.

        Returns:
            bool: True if a match was found, False otherwise.

        """

        if content_type is None:
            return False

        if content_type in available_content_types:
            return True

        for available_content_type in available_content_types:
            if available_content_type in content_type:
                return True

        return False


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.Crawler import CrawlerActions

import requests

class Options:
    """The Options class contains all the crawling settings/options.

    Attributes:
        scope (obj): Can be used to define the crawling scope.
        callbacks (obj): Can be used to define crawling callbacks.
        performance (obj): Can be used to define performance options.
        identity (obj): Can be used to define the identity/footprint options.

    """

    scope = None

    callbacks = None

    performance = None

    identity = None

    def __init__(self):
        """Constructs an Options class."""

        self.scope = OptionsScope()
        self.callbacks = OptionsCallbacks()
        self.performance = OptionsPerformance()
        self.identity = OptionsIdentity()

class OptionsScope:
    """The OptionsScope class contains the scope settings/options.

    Attributes:
        protocol_must_match (bool): only crawl pages with the same protocol as the startpoint (e.g. only https).
        subdomain_must_match (bool): only crawl pages with the same subdomain as the startpoint, if the startpoint is not a subdomain, no subdomains will be crawled.
        domain_must_match (bool): only crawl pages with the same domain as the startpoint (e.g. only finnwea.com).
        ignore_similar_requests (bool): ignore similar requests like `?page=1` & `?page=2` or `/page/1` and `/page/2`.
        max_depth (obj): the maximum search depth. For example, 2 would be the startpoint and all the pages found on it. Default is None (unlimited).

    """

    protocol_must_match = False

    subdomain_must_match = True

    domain_must_match = True

    ignore_similar_requests = False

    max_depth = None

class OptionsCallbacks:
    """The OptionsCallbacks class contains all the callback methods.

    Attributes:
        crawler_before_start (obj): called when the crawler starts crawling. Default is a null route.
        crawler_after_finish (obj): called when the crawler finished crawling. Default is a null route.
        request_before_start (obj): called when the crawler starts a new request. Default is a null route.
        request_after_finish (obj): called when the crawler finishes a request. Default is a null route.

    """

    crawler_before_start = None

    crawler_after_finish = None

    request_before_start = None

    request_after_finish = None

    def __init__(self):
        """Constructs an OptionsCallbacks class."""

        self.crawler_before_start = self.__null_route_crawler_before_start
        self.crawler_after_finish = self.__null_route_crawler_after_finish
        self.request_before_start = self.__null_route_request_before_start
        self.request_after_finish = self.__null_route_request_after_finish

    def __null_route_crawler_before_start(self):
        """A null route for the 'crawler before start' callback."""

        pass

    def __null_route_crawler_after_finish(self, queue):
        """A null route for the 'crawler after finish' callback.

        Args:
            queue (obj): The current crawling queue.

        """

        pass

    def __null_route_request_before_start(self, queue_item):
        """A null route for the 'request before start' callback.

        Args:
            queue_item (obj): The queue item that's about to start.

        Returns:
            str: A crawler action (either DO_SKIP_TO_NEXT, DO_STOP_CRAWLING or DO_CONTINUE_CRAWLING).

        """

        return CrawlerActions.DO_CONTINUE_CRAWLING

    def __null_route_request_after_finish(self, queue_item, new_queue_items):
        """A null route for the 'request after finish' callback.

        Args:
            queue_item (obj): The queue item that was finished.
            new_queue_items (obj): The new queue items that were found in the one that finished.

        Returns:
            str: A crawler action (either DO_STOP_CRAWLING or DO_CONTINUE_CRAWLING).

        """

        return CrawlerActions.DO_CONTINUE_CRAWLING

class OptionsPerformance:
    """The OptionsPerformance class contains the performance settings/options.

    Attributes:
        max_threads (obj): the maximum amount of simultaneous threads to use for crawling.
        
    """

    max_threads = 4

class OptionsIdentity:
    """The OptionsIdentity class contains the identity/footprint settings/options.

    Attributes:
        user_agent (str): the user agent to use when making a request.
        cookies (obj): the cookies to use when making a request.
        
    """

    # ToDo: Default this string to a random user agent
    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36"

    cookies = requests.cookies.RequestsCookieJar()


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.scrapers.XMLLinkScraper import XMLLinkScraper
from nyawc.Queue import Queue, QueueItem
from nyawc.http.Request import Request
from nyawc.http.Response import Response

import unittest
 
class TestScrapersXMLLinkScraper(unittest.TestCase):
    """The TestScrapersXMLLinkScraper tests if the XMLLinkScraper is working correctly.

    Attributes:
        __host str: The host were the new URLs were found on
        __urls list(obj): The URLs that were found

    """

    __host = "https://example.ltd/"

    __urls = [
        {"url": """https://example.ltd/?unique=1""", "must_pass": True, "test": """<link>https://example.ltd/?unique=1</link>"""},
        {"url": """http://example.ltd/?unique=2""", "must_pass": True, "test": """<link>http://example.ltd/?unique=2</link>"""},
        {"url": """https://example.ltd/?unique=3""", "must_pass": True, "test": """<link>//example.ltd/?unique=3</link>"""},
        {"url": """https://example.ltd/aa/bb/?unique=4""", "must_pass": True, "test": """<link>/aa/bb/?unique=4</link>"""},

        {"url": None, "must_pass": False, "test": """<abc>/aa/bb/?unique=4</def>"""},
        {"url": None, "must_pass": False, "test": """<link>asdfasdf/asdfasdf</link>"""},
    ]

    def test_xml_url_count(self):
        """Test if the amount of URLs found complies with the expected amount."""

        html = ""
        for url in self.__urls:
            html += "\n" + url["test"]

        finder = XMLLinkScraper(QueueItem(Request(""), Response()))
        matches = finder.get_requests_from_content(self.__host, html)

        self.assertEqual(len(matches), 4)
 
    def test_xml_url_matches(self):
        """Test if all the URLs match the found URLs."""
        
        for url in self.__urls:
            finder = XMLLinkScraper(QueueItem(Request(""), Response()))
            requests = finder.get_requests_from_content(self.__host, url["test"])

            if url["must_pass"]:
                self.assertEqual(requests[0].url, url["url"])
                self.assertEqual(len(requests), 1)
            else:
                self.assertEqual(len(requests), 0)



# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.http.Request import Request
from nyawc.helpers.URLHelper import URLHelper
from bs4 import BeautifulSoup

import html5lib

class SoupFormScraper:
    """The SoupLinkScraper finds requests from forms in HTML using BeautifulSoup.

    Attributes:
        content_types list(str): The supported content types.
        __queue_item (obj): The queue item containing the response to scrape.

    """

    content_types = [
        "text/html"
    ]

    __queue_item = None

    def __init__(self, queue_item):
        """Construct the SoupFormScraper class.

        Args:
            queue_item (obj): The queue item containing a response the scrape.

        """

        self.__queue_item = queue_item

    def get_requests(self):
        """Get all the new requests that were found in the response.

        Returns:
            list(obj): A list of new requests.

        """

        host = self.__queue_item.request.url
        content = self.__queue_item.response.text

        return self.get_requests_from_content(host, content)

    def get_requests_from_content(self, host, content):
        """Find new requests from the given content.

        Args:
            host (str): The parent request URL.
            content (obj): The HTML content.

        Returns:
            list(obj): Requests that were found.

        """

        soup = BeautifulSoup(content, "html5lib")
        forms = soup.find_all("form")

        found_requests = []

        for form in forms:
            found_requests.append(self.__get_request(host, form))

        return found_requests

    def __get_request(self, host, soup):
        """Build a request from the given soup form.

        Args:
            host str: The URL of the current queue item.
            soup (obj): The BeautifulSoup form.

        Returns:
            obj: The new Request.

        """

        url = URLHelper.make_absolute(host, self.__trim_grave_accent(soup['action'])) if soup.has_attr('action') else host
        method_original = soup['method'] if soup.has_attr('method') else 'get'
        method = 'post' if method_original.lower() == 'post' else 'get'
        data = self.__get_form_data(soup)

        return Request(url, method, data)


    def __trim_grave_accent(self, href):
        """Trim grave accents manually (because BeautifulSoup doesn't support it).

        Args:
            href (str): The BeautifulSoup href value.

        Returns:
            str: The BeautifulSoup href value without grave accents.

        """

        if href.startswith("`"):
            href = href[1:]

        if href.endswith("`"):
            href = href[:-1]

        return href

    def __get_form_data(self, soup):
        """Build a form data dict from the given form.

        Args:
            soup (obj): The BeautifulSoup form.

        Returns:
            obj: The form data (key/value).

        """

        fields = {}

        self.__get_form_data_from_inputs(soup, fields)
        self.__get_form_data_from_buttons(soup, fields)
        self.__get_form_data_from_textareas(soup, fields)
        self.__get_form_data_from_selects(soup, fields)

        return fields

    def __get_form_data_from_inputs(self, soup, fields):
        """Parse all the form data from input elements

        Args:
            soup (obj): The BeautifulSoup form.
            fields (obj): The fields (key/value) dict.

        """

        for input in soup.find_all('input'):
            if not input.has_attr('name'):
                continue
            
            if input['type'] in ('text', 'hidden', 'email', 'password', 'submit', 'image', 'color', 'date', 'month', 'number', 'search', 'tel', 'time', 'url', 'week', 'range'):
                fields[input['name']] = input['value'] if input.has_attr('value') else ''
                continue
            
            if input['type'] in ('checkbox', 'radio'):
                value = ''

                if input.has_attr('checked'):
                    if input.has_attr('value'):
                        value = input['value']
                    else:
                        value = 'on'

                if input['name'] in fields and value:
                    fields[input['name']] = value
                
                if not input['name'] in fields:
                    fields[input['name']] = value
                
                continue

    def __get_form_data_from_buttons(self, soup, fields):
        """Parse all the form data from button elements

        Args:
            soup (obj): The BeautifulSoup form.
            fields (obj): The fields (key/value) dict.

        """

        for button in soup.find_all('button'):
            if not button.has_attr('name'):
                continue

            fields[button['name']] = button['value'] if button.has_attr('value') else ''

    def __get_form_data_from_textareas(self, soup, fields):
        """Parse all the form data from textarea elements

        Args:
            soup (obj): The BeautifulSoup form.
            fields (obj): The fields (key/value) dict.

        """

        for textarea in soup.find_all('textarea'):
            if not textarea.has_attr('name'):
                continue

            fields[textarea['name']] = textarea.string or ''

    def __get_form_data_from_selects(self, soup, fields):
        """Parse all the form data from select elements

        Args:
            soup (obj): The BeautifulSoup form.
            fields (obj): The fields (key/value) dict.

        """

        for select in soup.find_all('select'):
            if not select.has_attr('name'):
                continue

            value = ''

            options = select.find_all('option')
            is_multiple = select.has_attr('multiple')

            selected_options = [
                option for option in options
                if option.has_attr('selected')
            ]
            
            if not selected_options and options:
                selected_options = [options[0]]
            
            if not is_multiple:
                if len(selected_options) >= 1:
                    value = selected_options[0]['value']
            else:
                value = [option['value'] for option in selected_options]
            
            fields[select['name']] = value


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.http.Request import Request
from nyawc.http.Response import Response

class Queue:
    """The Queue class contains all the request/response pairs that are going to be or have been crawled.

    Attributes:
        __items list(obj): The request/response pairs (as QueueItem's). 

    """

    __items = []

    def add(self, item):
        """Add a request/response pair (QueueItem) to the queue.

        Args:
            item (obj): The QueueItem to add.

        """

        self.__items.append(item)

    def add_request(self, request):
        """Add a request to the queue.

        Args:
            request (obj): The request to add.

        Returns:
            obj: The created queue item.

        """

        queue_item = QueueItem(request, Response())

        self.add(queue_item)
        
        return queue_item

    def get_first(self, status):
        """Get the first item in the queue that has the given status.

        Args:
            status (str): return the first item with this status.

        Returns:
            obj: The first queue item with the given status.

        """

        for item in self.__items:
            if item.status is status:
                return item

        return None

    def get_all(self):
        """Get all the items in the queue.

        Returns:
            list(obj): All the queue items.

        """

        return self.__items

    def get_all_including(self, include):
        """Get all the items that contain atleast one of the given statuses.

        Args:
            include list(str): an array of statuses.

        Returns:
            list(obj): All the queue items with one of the given statuses.

        """

        filtered_items = []

        for item in self.__items:
            if item.status in include:
                filtered_items.append(item)

        return filtered_items

    def get_all_excluding(self, exclude):
        """Get all the items that do not contain one of the given statuses.

        Args:
            exclude list(str): an array of statuses.

        Returns:
            list(obj): All the queue items that don't have one of the given statuses.

        """

        filtered_items = []

        for item in self.__items:
            if item.status not in exclude:
                filtered_items.append(item)

        return filtered_items

    def get_count(self):
        """Get a count of all the items in the queue."""

        return len(self.__items)

    def get_count_including(self, include):
        """Get a count of the items that contain atleast one of the given statuses.

        Args:
            include list(str): an array of statuses.

        Returns:
            int: The amount of queue items that have one of the given statuses.

        """

        count = 0

        for item in self.__items:
            if item.status in include:
                count += 1

        return count

    def get_count_excluding(self, exclude):
        """Get a count of the items that do not contain one of the given statuses.

        Args:
            exclude list(str): an array of statuses.

        Returns:
            list(obj): The amount of queue items that don't have one of the given statuses.

        """

        count = 0

        for item in self.__items:
            if item.status not in exclude:
                count += 1

        return count

    def get_progress(self):
        """Get the progress of the queue in percentage (float).

        Returns:
            float: The 'finished' progress in percentage.

        """

        all_count = self.get_count()
        finished_count = self.get_count_excluding([
            QueueItem.STATUS_QUEUED, 
            QueueItem.STATUS_IN_PROGRESS
        ])

        return 100 / all_count * finished_count

class QueueItem:
    """The QueueItem class keeps track of the request, response and the crawling status.

    Attributes:
        STATUS_QUEUED (str): Status for when the crawler did not yet start the request.
        STATUS_IN_PROGRESS (str): Status for when the crawler is currently crawling the request.
        STATUS_FINISHED (str): Status for when the crawler has finished crawling the request.
        STATUS_CANCELLED (str): Status for when the crawler has cancelled the request.
        STATUS_ERRORED (str): Status for when the crawler could not execute the request.
        status (str): The current crawling status.
        request (obj): The Request object.
        response (obj): The Response object.

    """

    STATUS_QUEUED = "queued"

    STATUS_IN_PROGRESS = "in_progress"

    STATUS_FINISHED = "finished"

    STATUS_CANCELLED = "cancelled"

    STATUS_ERRORED = "errored"

    status = STATUS_QUEUED

    request = None

    response = None

    def __init__(self, request, response):
        """Constructs a QueueItem class.

        Args:
            request (obj): The Request object.
            response (obj): The Response object (empty object when initialized).

        """
        
        self.request = request
        self.response = response


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.scrapers.SoupLinkScraper import SoupLinkScraper
from nyawc.Queue import Queue, QueueItem
from nyawc.http.Request import Request
from nyawc.http.Response import Response

import unittest
 
class TestScrapersSoupLinkScraper(unittest.TestCase):
    """The TestScrapersSoupLinkScraper tests if the SoupLinkScraper is working correctly.

    Attributes:
        __host str: The host were the new URLs were found on
        __urls list(obj): The URLs that were found

    """

    __host = "https://example.ltd/"

    __urls = [
        {"url": """https://example.ltd?unique=1""", "must_pass": True, "test": """<a href="https://example.ltd?unique=1">test</a>"""},
        {"url": """https://example.ltd/?unique=2""", "must_pass": True, "test": """<a href="https://example.ltd/?unique=2">test</a>"""},
        {"url": """http://example.ltd?unique=3""", "must_pass": True, "test": """<a href="http://example.ltd?unique=3">test</a>"""},
        {"url": """http://example.ltd/?unique=4""", "must_pass": True, "test": """<a href="http://example.ltd/?unique=4">test</a>"""},
        {"url": """https://example.ltd?unique=5""", "must_pass": True, "test": """<a href="//example.ltd?unique=5">test</a>"""},
        {"url": """https://example.ltd/?unique=6""", "must_pass": True, "test": """<a href="//example.ltd/?unique=6">test</a>"""},
        {"url": """https://example.ltd/?unique=7""", "must_pass": True, "test": """<a a="b" c=d href="//example.ltd/?unique=7">test</a>"""},
        {"url": """https://example.ltd/?unique=8""", "must_pass": True, "test": """<a href="//example.ltd/?unique=8" a=b c="d">test</a>"""},
        {"url": """https://example.ltd/?unique=9""", "must_pass": True, "test": """<a a="b" c=d href="//example.ltd/?unique=9" a=b c="d">test</a>"""},
        {"url": """https://example.ltd/index.php?unique=10""", "must_pass": True, "test": """<a href="https://example.ltd/index.php?unique=10">test</a>"""},
        {"url": """https://example.ltd/index.php?unique=11&d=c""", "must_pass": True, "test": """<a href="https://example.ltd/index.php?unique=11&d=c">test</a>"""},
        {"url": """https://example.ltd/index.php?unique=12&utf8=✓""", "must_pass": True, "test": """<a href="https://example.ltd/index.php?unique=12&utf8=✓">test</a>"""},
        {"url": """https://example.ltd/index.php?unique=13#anchor""", "must_pass": True, "test": """<a href="https://example.ltd/index.php?unique=13#anchor">test</a>"""},
        {"url": """https://example.ltd/folder1/folder2/folder3?unique=14""", "must_pass": True, "test": """<a href="https://example.ltd/folder1/folder2/folder3?unique=14">test</a>"""},
        {"url": """https://example.ltd/folder1/../folder2/folder3?unique=15""", "must_pass": True, "test": """<a href="https://example.ltd/folder1/../folder2/folder3?unique=15">test</a>"""},
        {"url": """https://example.ltd/../folder1/folder2/folder3?unique=16""", "must_pass": True, "test": """<a href="https://example.ltd/../folder1/folder2/folder3?unique=16">test</a>"""},
        {"url": """https://example.ltd/folder1/folder2/folder3?unique=17""", "must_pass": True, "test": """<a href="/folder1/folder2/folder3?unique=17">test</a>"""},
        {"url": """https://example.ltd/folder1/folder2/folder3?unique=18""", "must_pass": True, "test": """<a href="../folder1/folder2/folder3?unique=18">test</a>"""},
        {"url": """https://example.ltd/folder1/folder2/folder3?unique=19""", "must_pass": True, "test": """<a href="../../folder1/folder2/folder3?unique=19">test</a>"""},
        {"url": """https://example.ltd/folder1/folder2/folder3?unique=20""", "must_pass": True, "test": """<a href="/../../folder1/folder2/folder3?unique=20">test</a>"""},
        {"url": """https://example.ltd/?unique=21""", "must_pass": True, "test": """<a href='https://example.ltd/?unique=21'>test</a>"""},
        {"url": """https://example.ltd/?unique=22""", "must_pass": True, "test": """<a href=`https://example.ltd/?unique=22`>test</a>"""},
        {"url": """https://example.ltd/unique=23/folder'/?unique=23""", "must_pass": True, "test": """<a href=`https://example.ltd/unique=23/folder'/?unique=23`>test</a>"""},
        {"url": """https://example.ltd/unique=24/folder"/?unique=24""", "must_pass": True, "test": """<a href=`https://example.ltd/unique=24/folder"/?unique=24`>test</a>"""},
        {"url": """https://example.ltd/unique=25/folder'/?unique=25""", "must_pass": True, "test": """<a href="https://example.ltd/unique=25/folder'/?unique=25">test</a>"""},
        {"url": """https://example.ltd/unique=26/folder`/?unique=26""", "must_pass": True, "test": """<a href="https://example.ltd/unique=26/folder`/?unique=26">test</a>"""},
        {"url": """https://example.ltd/unique=27/folder"/?unique=27""", "must_pass": True, "test": """<a href='https://example.ltd/unique=27/folder"/?unique=27'>test</a>"""},
        {"url": """https://example.ltd/unique=28/folder`/?unique=28""", "must_pass": True, "test": """<a href='https://example.ltd/unique=28/folder`/?unique=28'>test</a>"""},
        {"url": """https://example.ltd/unique=29/folder`/?unique=29""", "must_pass": True, "test": """<a href='https://example.ltd/unique=29/folder`/?unique=29'&b=not-included'>test</a>"""},
        {"url": """https://example.ltd/unique=30/folder`/?unique=30'&b=included""", "must_pass": True, "test": """<a href="https://example.ltd/unique=30/folder`/?unique=30'&b=included">test</a>"""},
        {"url": """https://example.ltd/sample/page/1?unique=31""", "must_pass": True, "test": """<a href="sample/page/1?unique=31">test</a>"""},
        {"url": """https://example.ltd/page/2?unique=32""", "must_pass": True, "test": """<a href="/page/2?unique=32">test</a>"""},
        {"url": """https://example.ltd/page/3?unique=33""", "must_pass": True, "test": """<a href="page/3?unique=33">test</a>"""},
        {"url": """https://example.ltd/page4?unique=34""", "must_pass": True, "test": """<a href="page4?unique=34">test</a>"""},
        {"url": """http://examp%0ale.ltd/?unique=35""", "must_pass": True, "test": """<a href="http://examp%0ale.ltd/?unique=35">"""},
        {"url": """http://examp\nle.ltd/?unique=36""", "must_pass": True, "test": """<a href="http://examp\nle.ltd/?unique=36">"""}
    ]

    def test_soup_url_count(self):
        """Test if the amount of URLs found complies with the expected amount."""

        html = ""
        for url in self.__urls:
            html += "\n" + url["test"]

        finder = SoupLinkScraper(QueueItem(Request(""), Response()))
        matches = finder.get_requests_from_content(self.__host, html)

        self.assertEqual(len(matches), 36)
 
    def test_soup_url_matches(self):
        """Test if all the URLs match the found URLs."""

        for url in self.__urls:
            finder = SoupLinkScraper(QueueItem(Request(""), Response()))
            requests = finder.get_requests_from_content(self.__host, url["test"])

            if url["must_pass"]:
                self.assertEqual(requests[0].url, url["url"])
                self.assertEqual(len(requests), 1)
            else:
                self.assertEqual(len(requests), 0)



# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.helpers.URLHelper import URLHelper
from nyawc.helpers.CookieHelper import CookieHelper

class Request:
    """The Request class contains details that were used to request the specified URL.

    Attributes:
        REQUEST_METHOD_OPTIONS (str): A request method that can be used to request the URL.
        REQUEST_METHOD_GET (str): A request method that can be used to request the URL.
        REQUEST_METHOD_HEAD (str): A request method that can be used to request the URL.
        REQUEST_METHOD_POST (str): A request method that can be used to request the URL.
        REQUEST_METHOD_PUT (str): A request method that can be used to request the URL.
        REQUEST_METHOD_DELETE (str): A request method that can be used to request the URL.
        REQUEST_METHODS list(str): All available request methods in a list.
        parent_raised_error (bool): If the parent request raised an error (e.g. 404).
        depth (int): The current crawling depth.
        url (str): The absolute URL to use when making the request.
        method (str): The request method to use for the request.
        data (obj): The post data {key: value} object that will be sent.
        cookies (obj): The (requests module) cookie jar to use for the request.
        user_agent (str): The user agent to use for the request.

    """

    METHOD_OPTIONS = "options"

    METHOD_GET = "get"

    METHOD_HEAD = "head"

    METHOD_POST = "post"

    METHOD_PUT = "put"

    METHOD_DELETE = "delete"

    METHODS = [
        METHOD_OPTIONS,
        METHOD_GET,
        METHOD_HEAD,
        METHOD_POST,
        METHOD_PUT,
        METHOD_DELETE
    ]

    parent_raised_error = False

    depth = 0

    url = None

    method = None

    data = None

    cookies = None

    user_agent = None

    def __init__(self, url, method=METHOD_GET, data=None, cookies=None, user_agent=None):
        """Constructs a Request class.

        Args:
            url (str): The absolute URL to use when making the request.
            method (str): The request method to use for the request.
            data (obj): The post data {key: value} object that will be sent.
            cookies (obj): The (requests module) cookie jar to use for the request.
            user_agent (str): The user agent to use for the request.

        """

        self.url = url
        self.method = method
        self.cookies = cookies
        self.user_agent = user_agent

        if method == self.METHOD_GET:
            self.url = URLHelper.append_with_data(self.url, data)
        else:
            self.data = data

    def is_same_as(self, request):
        """Check if this request is the same to the given request.

        Args:
            request (obj): Check if self is the same as this request.

        Returns:
            bool: True if the same, False otherwise.

        """

        if self.url != request.url:
            return False

        if self.method != request.method:
            return False

        if not CookieHelper.are_cookies_the_same(self.cookies, request.cookies):
            return False

        return True

    def is_similar_to(self, request):
        """Check if this request is similar to the given request.

        Args:
            request (obj): Check if self is similar as this request.

        Returns:
            bool: True if similar, False otherwise.

        """

        if self.method != request.method:
            return False

        if not URLHelper.is_data_similar(self.data, request.data):
            return False

        if not CookieHelper.are_cookies_similar(self.cookies, request.cookies):
            return False

        if self.url != request.url:
            if not URLHelper.are_urls_similar(self.url, request.url):
                return False

        return True


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.http.Request import Request
from nyawc.helpers.URLHelper import URLHelper

import re

class XMLLinkScraper:
    """The XMLLinkScraper finds absolute URLs in XML values.

    Attributes:
        content_types list(str): The supported content types.
        __expressions list(obj): The queue item containing the response to scrape.
        __queue_item (obj): The queue item containing the response to scrape.

    """

    content_types = [
        "text/xml",
        "application/xml",
    ]

    __expressions = [
        # Match absolute/relative URLs between any type of HTML quote
        { "group": 2, "raw": r"(\<(.*)\>)(((((https?:)?\/)?\/)|(\.\.\/)+)(.*?))(\<\/\2\>)" },
    ]

    __queue_item = None

    def __init__(self, queue_item):
        """Construct the XMLLinkScraper class.

        Args:
            queue_item (obj): The queue item containing a response the scrape.

        """

        self.__queue_item = queue_item

    def get_requests(self):
        """Get all the new requests that were found in the response.

        Returns:
            list(obj): A list of new requests.

        """

        host = self.__queue_item.request.url
        content = self.__queue_item.response.text

        return self.get_requests_from_content(host, content)

    def get_requests_from_content(self, host, content):
        """Find new requests from the given content.

        Args:
            host (str): The parent request URL.
            content (obj): The HTML content.

        Returns:
            list(obj): Requests that were found.

        """

        found_requests = []

        for expression in self.__expressions:
            matches = re.findall(expression["raw"], content)

            for match in matches:
                found_url = match[expression["group"]]
                absolute_url = URLHelper.make_absolute(host, found_url)
                found_requests.append(Request(absolute_url))

        return found_requests


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from nyawc.http.Request import Request
from nyawc.helpers.URLHelper import URLHelper
from bs4 import BeautifulSoup

import html5lib

class SoupLinkScraper:
    """The SoupLinkScraper finds URLs from href attributes in HTML using BeautifulSoup.

    Attributes:
        content_types list(str): The supported content types.
        __queue_item (obj): The queue item containing the response to scrape.

    """

    content_types = [
        "text/html"
    ]

    __queue_item = None

    def __init__(self, queue_item):
        """Construct the SoupLinkScraper class.

        Args:
            queue_item (obj): The queue item containing a response the scrape.

        """

        self.__queue_item = queue_item

    def get_requests(self):
        """Get all the new requests that were found in the response.

        Returns:
            list(obj): A list of new requests.

        """

        host = self.__queue_item.request.url
        content = self.__queue_item.response.text

        return self.get_requests_from_content(host, content)
        

    def get_requests_from_content(self, host, content):
        """Find new requests from the given content.

        Args:
            host (str): The parent request URL.
            content (obj): The HTML content.

        Returns:
            list(obj): Requests that were found.

        """

        soup = BeautifulSoup(content, "html5lib")
        a_elements = soup.find_all("a", href=True)
        link_elements = soup.find_all("link", href=True)
        script_elements = soup.find_all("script", src=True)

        found_requests = []

        for a_element in a_elements:
            found_url = self.__trim_grave_accent(a_element["href"])
            if not URLHelper.is_mailto(found_url):
                absolute_url = URLHelper.make_absolute(host, found_url)
                found_requests.append(Request(absolute_url))

        for link_element in link_elements:
            found_url = self.__trim_grave_accent(link_element["href"])
            if not URLHelper.is_mailto(found_url):
                absolute_url = URLHelper.make_absolute(host, found_url)
                found_requests.append(Request(absolute_url))

        for script_element in script_elements:
            found_url = self.__trim_grave_accent(script_element["src"])
            if not URLHelper.is_mailto(found_url):
                absolute_url = URLHelper.make_absolute(host, found_url)
                found_requests.append(Request(absolute_url))

        return found_requests

    def __trim_grave_accent(self, href):
        """Trim grave accents manually (because BeautifulSoup doesn't support it).

        Args:
            href (str): The BeautifulSoup href value.

        Returns:
            str: The BeautifulSoup href value without grave accents.

        """

        if href.startswith("`"):
            href = href[1:]

        if href.endswith("`"):
            href = href[:-1]

        return href


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from urllib.parse import urlparse

class HTTPRequestHelper:
    """A helper for the src.http.Request module."""

    @staticmethod
    def patch_with_options(request, options, parent_response=None):
        """Patch the given request with the given options (e.g. user agent).

        Args:
            request (obj): The request to check.
            options (obj): The options to patch the request with.
            parent_response(obj): The parent response object if exists.

        """

        request.user_agent = options.identity.user_agent
        request.cookies = options.identity.cookies

        if parent_response != None:
            for cookie in parent_response.cookies:
                request.cookies.set(cookie.name, cookie.value, domain=cookie.domain, path=cookie.path)

    @staticmethod
    def is_already_in_queue(request, queue):
        """Check if the given request is already in the queue.

        Args:
            request (obj): The request to check.
            queue (obj): The queue to look in.

        Returns:
            bool: True if in queue, False otherwise.

        """

        for queue_item in queue.get_all():
            if queue_item.request.is_same_as(request):
                return True

        return False

    @staticmethod
    def is_similar_already_in_queue(request, queue):
        """Check if the given request (or a similar one) is already in the queue.

        Args:
            request (obj): The request to check.
            queue (obj): The queue to look in.

        Returns:
            bool: True if in queue, False otherwise.

        """

        for queue_item in queue.get_all():
            if queue_item.request.is_similar_to(request):
                return True

        return False

    @staticmethod
    def complies_with_scope(queue, queue_item, new_request, scope):
        """Check if the new requests complies with the crawling scope.

        Args:
            queue (obj): The current crawling queue.
            queue_item (obj): The parent queue item of the new request.
            new_request (obj): The request to check.
            scope (obj): The scope to check.

        Returns:
            bool: True if it complies, False otherwise.

        """

        try:
            parsed_url = urlparse(queue_item.request.url)
            parsed_new_url = urlparse(new_request.url)
        except:
            return False

        subdomain = parsed_url.netloc.split(".")[:-2]
        subdomain_new = parsed_new_url.netloc.split(".")[:-2]

        if scope.protocol_must_match:
            if parsed_url.scheme != parsed_new_url.scheme:
                return False

        if scope.subdomain_must_match:
            if subdomain != subdomain_new:
                return False

        if scope.domain_must_match:
            if parsed_url.netloc != parsed_new_url.netloc:
                return False

        if scope.ignore_similar_requests:
            if HTTPRequestHelper.is_similar_already_in_queue(new_request, queue):
                return False

        return True


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import find_packages, setup

with open('requirements.txt') as file:
    requirements = file.read().splitlines()

setup(
    version="1.2.1",
    name="nyawc",
    keywords = "web-crawler web-scraper request-crawler vulnerability-scanner crawler scanner",
    description="A web crawler that gathers requests instead of URLs.",
    long_description="A very useful web crawler for vulnerability scanning. Not Your Average Web Crawler (N.Y.A.W.C) is a Python application that enables you to crawl web applications for requests instead of URLs. It crawls every GET and POST request on the specified domain and keeps track of the request and response data. It's main purpose is to be used in web application vulnerability scanners",
    packages=find_packages(),
    platforms=["any"],
    author="Tijme Gommers",
    author_email="tijme@finnwea.com",
    license="MIT",
    url="https://github.com/tijme/not-your-average-web-crawler",
    install_requires=requirements
)


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class CookieHelper:
    """A helper for cookie jars."""

    @staticmethod
    def are_cookies_the_same(jar1, jar2):
        """Check if the given cookies are the same.

        Args:
            jar1 (str): The first cookie jar.
            jar1 (str): The second cookie jar.

        Returns:
            bool: If the cookies are the same

        """


        if jar1 is None and jar2 is not None:
            return False

        if jar2 is None and jar1 is not None:
            return False

        if jar1 is None and jar2 is None:
            return True

        cookieMatches = False
        for cookie1 in jar1:
            for cookie2 in jar2:
                if cookie1.name == cookie2.name:
                    if cookie1.value == cookie2.value:
                        cookieMatches = True

        if not cookieMatches:
            return False

        cookieMatches = False
        for cookie2 in jar2:
            for cookie1 in jar1:
                if cookie2.name == cookie1.name:
                    if cookie2.value == cookie1.value:
                        cookieMatches = True

        if not cookieMatches:
            return False
                
        return True

    @staticmethod
    def are_cookies_similar(jar1, jar2):
        """Check if the given cookies are similar.

        Args:
            jar1 (str): The first cookie jar.
            jar1 (str): The second cookie jar.

        Returns:
            bool: If the cookies are similar

        """

        return True

        if jar1 is None and jar2 is not None:
            return False

        if jar2 is None and jar1 is not None:
            return False

        if jar1 is None and jar2 is None:
            return True

        for cookie1 in jar1:
            cookieMatches = False
            for cookie2 in jar2:
                if cookie1.name == cookie2.name:
                    cookieMatches = True

        if not cookieMatches:
            return False

        for cookie2 in jar2:
            cookieMatches = False
            for cookie1 in jar1:
                if cookie2.name == cookie1.name:
                    cookieMatches = True

        if not cookieMatches:
            return False
                
        return True


# -*- coding: utf-8 -*-

# MIT License
# 
# Copyright (c) 2017 Tijme Gommers
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from urllib.parse import urljoin, urlparse, parse_qsl, urlencode, urlunparse

class URLHelper:
    """A helper for URL strings."""

    @staticmethod
    def make_absolute(parent_absolute, relative):
        """Make the given (relative) URL absolute.

        Args:
            parent_absolute (str): The absolute URL the relative url was found on.
            relative (str): The (possibly relative) url to make absolute.

        Returns:
            str: The absolute URL.

        """

        if relative.startswith("http://") or relative.startswith("https://"):
            return relative

        parsed_host = urlparse(parent_absolute)

        if relative.startswith("//"):
            return parsed_host.scheme + ":" + relative

        return urljoin(parent_absolute, relative)

    @staticmethod
    def append_with_data(url, data):
        """Append the given URL with the given data dict.

        Args:
            url (str): The URL to append.
            data (obj): The key value data to append to the URL.

        Returns:
            str: The new URL.

        """

        if data is None:
            return url

        url_parts = list(urlparse(url))

        query = dict(parse_qsl(url_parts[4]))
        query.update(data)

        url_parts[4] = urlencode(query)

        return urlunparse(url_parts)

    @staticmethod
    def are_urls_similar(url1, url2):
        """Check if the given URLs are similar to each other.

        Args:
            url1 (str): The first URL.
            url2 (str): The second URL.

        Returns:
            bool: True if similar, False otherwise.

        """

        parsed_url1 = urlparse(url1)
        parsed_url2 = urlparse(url2)

        if parsed_url1.netloc != parsed_url2.netloc:
            return False

        if parsed_url1.path != parsed_url2.path:
            return False

        dict_url1 = dict(parse_qsl(parsed_url1.query))
        dict_url2 = dict(parse_qsl(parsed_url2.query))

        return URLHelper.is_data_similar(dict_url1, dict_url2)

    @staticmethod
    def is_data_similar(data1, data2):
        """Check if the given data dicts are similar to each other.

        Args:
            data1 (obj): The first data object.
            data2 (obj): The second data object.

        Returns:
            bool: True if similar, False otherwise.

        """

        if data1 is None and data2 is not None:
            return False

        if data2 is None and data1 is not None:
            return False

        if data1 is None and data2 is None:
            return True

        return data1.keys() == data2.keys()

    @staticmethod
    def is_mailto(url):
        """Check if the given URL is a mailto URL

        Args:
            url (str): The URL to check.

        Returns:
            bool: True if mailto, False otherwise.

        """

        return url.startswith("mailto:")